<?php

/**
 * @file
 * Content creation / editing handler
 */

function drupad_content_forms() {
  switch (arg(2)) {
    case 'node':
      $data = drupad_content_get_form_for_node(arg(3));
      break;
    case 'node-add':
      $data = drupad_content_get_form_for_node_type(arg(3));
      break;
    case 'submit':
      $data = drupad_content_form_submit_end_point();
      break;
    case 'upload':
      $data = drupad_upload_js();
      break;
    default:
      $data = drupad_content_get_content_types();
      break;
  }
  drupad_serve_response($data);
}

/**
 * Get all content types the current user can create
 *
 * @return structured response array
 */
function drupad_content_get_content_types() {
  $types = node_get_types('types', NULL, FALSE);

  $content_types = array();
  $content_types_to_exclude = variable_get('drupad_content_types_excluded_node_add', array());

  foreach ($types as $key => $type) {
    if (node_access('create', $type->type) && !in_array($type->type, $content_types_to_exclude)) {
      $content_types[] = $type;
    }
  }

  $data['meta'] = array();
  $data['content'] = $content_types;

  return $data;
}

/**
 * We need to overwrite node_form_submit() to redirect node form
 * submissions to our own node view callback once submission succeed
 * @see drupad_node_form_submit()
 */
function drupad_form_alter(&$form, $form_state, $form_id) {
  // We only alter forms when the iOS app is requesting a node form
  if ($form['#id'] == 'node-form' && arg(0) == 'drupad' && arg(1) == 'content-forms') {
    $form['buttons']['submit']['#submit'][] = 'drupad_node_form_submit';
    /////////////////////////////////////////////////////
    // We add labels to upload module form components
    // since they are presented in a table when in HTML format
    if (module_exists('upload') && isset($form['attachments']['wrapper']['files'])) {
      // Upload module creates a draggable table. Labels are placed in that table
      // thus we need to re-set them to obtain something understandable on the app.
      foreach ($form['attachments']['wrapper']['files'] as $fid => $file_description) {
        if (is_array($file_description)) {
          $form['attachments']['wrapper']['files'][$fid]['description']['#title'] = t('Description');
          $form['attachments']['wrapper']['files'][$fid]['remove']['#title'] = t('Remove');
          $form['attachments']['wrapper']['files'][$fid]['remove']['#description'] = t('If checked, the file will be removed upon save <strong>permanently</strong>.');
          $form['attachments']['wrapper']['files'][$fid]['weight']['#title'] = t('Weight');
          $form['attachments']['wrapper']['files'][$fid]['list']['#title'] = t('List');
          $form['attachments']['wrapper']['files'][$fid]['list']['#description'] = t('If checked, the file will be listed on node view.');
        }
      }
    }
    /////////////////////////////////////////////////////
    // Remove the upload node submit button/handler
    if (module_exists('upload') && isset($form['attachments']['wrapper']['new']['attach'])) {
      unset($form['attachments']['wrapper']['new']['attach']);
    }
  }
}

/**
 * Set redirection of the node form
 *
 * @see node_form_submit()
 * @see drupad_form_alter()
 */
function drupad_node_form_submit($form, &$form_state) {
  // Redirect to the node edit form if it has succeed at saving
  if (isset($form_state['nid'])) {
    $form_state['redirect'] = 'drupad/content-forms/node/'. $form_state['nid'];
  }
}

function drupad_content_get_form_for_node_type($content_type) {
  module_load_include('inc', 'node', 'node.pages');
  $data = array();

  // We need to build the form before getting Errors information
  $data['content'] = drupad_node_add($content_type);
  $data['meta'] = array();
  $data['meta']['errors'] = form_get_errors() ? form_get_errors() : array();

  drupad_serve_response($data);
}

function drupad_content_get_form_for_node($nid) {
  module_load_include('inc', 'node', 'node.pages');
  $data = array('meta' => array());

  $node = node_load($nid);

  if (node_access('update', $node, NULL)) {
    // We need to build the form before getting Errors information
    $data['content'] = drupad_node_add($node->type, $node);
    $data['meta']['errors'] = form_get_errors() ? form_get_errors() : array();
  }
  else {
    $data['content'] = array();
    $data['meta']['errors'] = array('node_access' => FALSE);
    watchdog('access denied', '@q accessed via Drupad iOS app', array('@q' => $_GET['q']), WATCHDOG_WARNING);
  }

  drupad_serve_response($data);
}

/**
 * Returns a form structure either for a node type or a node
 *
 * @param $type
 *  the content type to retrieve
 * @param $node
 *  the node that should fill the form (can be NULL)
 *
 * @return either a form structured array or FALSE
 */
function drupad_node_add($type, $node = NULL) {
  global $user;

  $types      = node_get_types();
  $type       = isset($type) ? str_replace('-', '_', $type) : NULL;
  $form_id    = $type . '_node_form';
  $form_state = array();

  // If a node type has been specified, validate its existence.
  if (isset($types[$type]) && node_access('create', $type)) {
    // Initialize settings:
    if (!$node) {
      $node = array(
        'uid' => $user->uid,
        'name' => (isset($user->name) ? $user->name : ''),
        'type' => $type,
        'language' => '',
      );
    }

    $form = drupad_get_form_structure($type . '_node_form', $node);
    drupad_fix_node_form_admin_order($form);
    drupad_convert_form_options_to_strict_array($form);

    return $form;
  }

  return FALSE;
}

/**
 * Converts form component's options to strict array
 *
 * Walk through each form component recursively for #options array.
 * Converts #options array to a strict array (not a associative array)
 * in order to keep the order revelant, as hashes don't have specific order.
 *
 * Also replace 0 value as "0" (string) so that it isn't interpreted as BOOL value
 * by Objective-C.
 *
 * @param &$form
 */
function drupad_convert_form_options_to_strict_array(&$form) {
  $children_keys = element_children($form);
  foreach ($children_keys as $child_key) {
    // Reached an optionable component (select or radios)
    if (isset($form[$child_key]['#options']) && is_array($form[$child_key]['#options'])) {
      $options = array();
      // Replaces #options with a strict array since
      // hashes don't keep order outside of PHP
      foreach ($form[$child_key]['#options'] as $key => $value) {
        if (is_object($value)) {
          $option = $value->option;
          reset($option); // Shouldn't be useful
          $options[] = array(key($option) => current($option));
        }
        else {
          // Dirty hack, but no choice
          if ($value === 0 && $key === 0) {
            $key = "0";
          }
          $options[] = array((string) $key => $value);
        }
      }
      // Shiny new options ordered
      $form[$child_key]['#options'] = $options;
    }
    // Continue deeper in the form structure
    drupad_convert_form_options_to_strict_array($form[$child_key]);
  }
}

/**
 * Redefine all of the processing of drupal_get_form, without the rendering part
 *
 * We just want to send JSON formatted forms. Please note that due to
 * to the nature of iOS versus HTML rendering, some form components won't
 * feel, look and/or behave as you would expect. JavaScript additions won't be
 * be present, CSS used to format some component (ex: webform) won't have anymore effect.
 */
function drupad_get_form_structure($form_id) {
  $form_state = array('storage' => NULL, 'submitted' => FALSE);

  $args = func_get_args();
  $cacheable = FALSE;

  if (isset($_SESSION['batch_form_state'])) {
    // We've been redirected here after a batch processing : the form has
    // already been processed, so we grab the post-process $form_state value
    // and move on to form display. See _batch_finished() function.
    $form_state = $_SESSION['batch_form_state'];
    unset($_SESSION['batch_form_state']);
  }
  else {
    // If the incoming $_POST contains a form_build_id, we'll check the
    // cache for a copy of the form in question. If it's there, we don't
    // have to rebuild the form to proceed. In addition, if there is stored
    // form_state data from a previous step, we'll retrieve it so it can
    // be passed on to the form processing code.
    if (isset($_POST['form_id']) && $_POST['form_id'] == $form_id && !empty($_POST['form_build_id'])) {
      $form = form_get_cache($_POST['form_build_id'], $form_state);
    }

    // If the previous bit of code didn't result in a populated $form
    // object, we're hitting the form for the first time and we need
    // to build it from scratch.
    if (!isset($form)) {
      $form_state['post'] = $_POST;
      // Use a copy of the function's arguments for manipulation
      $args_temp = $args;
      $args_temp[0] = &$form_state;
      array_unshift($args_temp, $form_id);

      $form = call_user_func_array('drupal_retrieve_form', $args_temp);
      $form_build_id = 'form-'. md5(uniqid(mt_rand(), TRUE));
      $form['#build_id'] = $form_build_id;
      drupal_prepare_form($form_id, $form, $form_state);
      // Store a copy of the unprocessed form for caching and indicate that it
      // is cacheable if #cache will be set.
      $original_form = $form;
      $cacheable = TRUE;
      unset($form_state['post']);
    }
    $form['#post'] = $_POST;

    // Now that we know we have a form, we'll process it (validating,
    // submitting, and handling the results returned by its submission
    // handlers. Submit handlers accumulate data in the form_state by
    // altering the $form_state variable, which is passed into them by
    // reference.
    drupal_process_form($form_id, $form, $form_state);
    if ($cacheable && !empty($form['#cache'])) {
      // Caching is done past drupal_process_form so #process callbacks can
      // set #cache.
      form_set_cache($form_build_id, $original_form, $form_state);
    }
  }

  // Most simple, single-step forms will be finished by this point --
  // drupal_process_form() usually redirects to another page (or to
  // a 'fresh' copy of the form) once processing is complete. If one
  // of the form's handlers has set $form_state['redirect'] to FALSE,
  // the form will simply be re-rendered with the values still in its
  // fields.
  //
  // If $form_state['storage'] or $form_state['rebuild'] has been set
  // and input has been processed, we know that we're in a complex
  // multi-part process of some sort and the form's workflow is NOT
  // complete. We need to construct a fresh copy of the form, passing
  // in the latest $form_state in addition to any other variables passed
  // into drupal_get_form().

  if ((!empty($form_state['storage']) || !empty($form_state['rebuild'])) && !empty($form_state['process_input']) && !form_get_errors()) {
    $form = drupal_rebuild_form($form_id, $form_state, $args);
  }

  return drupad_render_form($form_id, $form);
}

function drupad_render_form($form_id, &$form) {
  // Don't override #theme if someone already set it.
  if (!isset($form['#theme'])) {
    init_theme();
    $registry = theme_get_registry();
    if (isset($registry[$form_id])) {
      $form['#theme'] = $form_id;
    }
  }

  return drupad_render($form);
}

function drupad_render(&$elements) {

  if (!isset($elements) || (isset($elements['#access']) && !$elements['#access'])) {
    return NULL;
  }

  // If the default values for this element haven't been loaded yet, populate
  // them.
  if (!isset($elements['#defaults_loaded']) || !$elements['#defaults_loaded']) {
    if ((!empty($elements['#type'])) && ($info = _element_info($elements['#type']))) {
      $elements += $info;
    }
  }

  // Make any final changes to the element before it is rendered. This means
  // that the $element or the children can be altered or corrected before the
  // element is rendered into the final text.
  if (isset($elements['#pre_render'])) {
    foreach ($elements['#pre_render'] as $function) {
      if (function_exists($function)) {
        $elements = $function($elements);
      }
    }
  }

  $content = array();
  // Either the elements did not go through form_builder or one of the children
  // has a #weight.
  if (!isset($elements['#sorted'])) {
    uasort($elements, "element_sort");
  }
  $elements += array('#title' => NULL, '#description' => NULL);
  if (!isset($elements['#children'])) {
    $children = element_children($elements);
    // Render all the children that use a theme function.
    if (isset($elements['#theme']) && empty($elements['#theme_used'])) {
      $elements['#theme_used'] = TRUE;

      $previous = array();
      foreach (array('#value', '#type', '#prefix', '#suffix') as $key) {
        $previous[$key] = isset($elements[$key]) ? $elements[$key] : NULL;
      }
      // If we rendered a single element, then we will skip the renderer.
      if (empty($children)) {
        $elements['#printed'] = TRUE;
      }
      else {
        $elements['#value'] = '';
      }
      $elements['#type'] = 'markup';

      //unset($elements['#prefix'], $elements['#suffix']);
      $content = $elements;

      foreach (array('#value', '#type', '#prefix', '#suffix') as $key) {
        $elements[$key] = isset($previous[$key]) ? $previous[$key] : NULL;
      }
    }
    // Render each of the children using drupal_render and concatenate them.
    if (count($content) == 0) {
      foreach ($children as $key) {
        $content[] = drupad_render($elements[$key]);
      }
    }
  }

  if (count($content) > 0) {
    $elements['#children'] = $content;
  }

  // // Until now, we rendered the children, here we render the element itself
  // if (!isset($elements['#printed'])) {
  //   $content = theme(!empty($elements['#type']) ? $elements['#type'] : 'markup', $elements);
  //   $elements['#printed'] = TRUE;
  // }

  if (count($content) > 0) {
    // Filter the outputted content and make any last changes before the
    // content is sent to the browser. The changes are made on $content
    // which allows the output'ed text to be filtered.
    if (isset($elements['#post_render'])) {
      foreach ($elements['#post_render'] as $function) {
        if (function_exists($function)) {
          $content = $function($content, $elements);
        }
      }
    }
    //$prefix = isset($elements['#prefix']) ? $elements['#prefix'] : '';
    //$suffix = isset($elements['#suffix']) ? $elements['#suffix'] : '';
    return $content;
  }
}

/**
 * Applies the same modification that theme_node_form about element ordering
 *
 * This allows us to present the UI form elements in the right order
 *
 * @see theme_node_form()
 */
function drupad_fix_node_form_admin_order(&$form) {
  if ($form['author']) {
    $form['author']['#weight'] = 99997;
  }
  if (isset($form['options'])) {
    $form['options']['#weight'] = 99998;
  }
  $form['buttons']['#weight'] = 99999;
}
